package dev.shadowsoffire.apotheosis.adventure.boss;

import com.mojang.datafixers.kinds.Applicative;
import dev.shadowsoffire.placebo.json.NBTAdapter;
import java.util.Collections;
import net.minecraftforge.registries.ForgeRegistries;
import net.minecraft.util.ExtraCodecs;
import dev.shadowsoffire.placebo.codec.PlaceboCodecs;
import com.mojang.datafixers.kinds.App;
import com.mojang.serialization.codecs.RecordCodecBuilder;
import net.minecraft.world.entity.MobSpawnType;
import net.minecraft.world.item.ItemStack;
import dev.shadowsoffire.apotheosis.Apotheosis;
import net.minecraft.network.chat.Style;
import net.minecraft.network.chat.MutableComponent;
import dev.shadowsoffire.apotheosis.adventure.loot.LootRarity;
import dev.shadowsoffire.apotheosis.adventure.AdventureConfig;
import dev.shadowsoffire.apotheosis.adventure.loot.RarityClamp;
import dev.shadowsoffire.apotheosis.adventure.AdventureModule;
import dev.shadowsoffire.apotheosis.adventure.loot.LootCategory;
import net.minecraft.world.entity.EquipmentSlot;
import com.google.common.base.Preconditions;
import dev.shadowsoffire.placebo.json.GearSetRegistry;
import net.minecraft.network.chat.Component;
import com.google.common.base.Strings;
import dev.shadowsoffire.apotheosis.util.NameHelper;
import net.minecraft.world.entity.LivingEntity;
import dev.shadowsoffire.placebo.json.RandomAttributeModifier;
import dev.shadowsoffire.placebo.json.ChancedEffectInstance;
import net.minecraft.world.entity.monster.Creeper;
import java.util.Iterator;
import net.minecraft.nbt.ListTag;
import net.minecraft.world.phys.Vec3;
import net.minecraft.world.entity.Entity;
import net.minecraft.world.level.Level;
import java.util.function.Function;
import net.minecraft.util.RandomSource;
import net.minecraft.world.entity.Mob;
import net.minecraft.world.level.ServerLevelAccessor;
import java.util.Optional;
import dev.shadowsoffire.apotheosis.util.SupportingEntity;
import net.minecraft.nbt.CompoundTag;
import dev.shadowsoffire.placebo.json.GearSet;
import java.util.List;
import net.minecraft.resources.ResourceLocation;
import javax.annotation.Nullable;
import net.minecraft.world.entity.EntityType;
import java.util.Set;
import com.mojang.serialization.Codec;
import dev.shadowsoffire.apotheosis.adventure.compat.GameStagesCompat;
import dev.shadowsoffire.placebo.reload.WeightedDynamicRegistry;
import dev.shadowsoffire.placebo.codec.CodecProvider;

public final class ApothMiniboss implements CodecProvider<ApothMiniboss>, WeightedDynamicRegistry.ILuckyWeighted, WeightedDynamicRegistry.IDimensional, GameStagesCompat.IStaged, MinibossRegistry.IEntityMatch
{
    public static final String NAME_GEN = "use_name_generation";
    public static final Codec<ApothMiniboss> CODEC;
    protected final int weight;
    protected final float quality;
    protected final float chance;
    protected final String name;
    protected final Set<EntityType<?>> entities;
    protected final BossStats stats;
    @Nullable
    protected final Set<String> stages;
    protected final Set<ResourceLocation> dimensions;
    protected final boolean affixed;
    protected final List<GearSet.SetPredicate> gearSets;
    @Nullable
    protected final CompoundTag nbt;
    protected final List<SupportingEntity> support;
    @Nullable
    protected final SupportingEntity mount;
    protected final List<Exclusion> exclusions;
    protected final boolean finalize;
    
    public ApothMiniboss(final int weight, final float quality, final float chance, final String name, final Set<EntityType<?>> entities, final BossStats stats, final Optional<Set<String>> stages, final Set<ResourceLocation> dimensions, final boolean affixed, final List<GearSet.SetPredicate> gearSets, final Optional<CompoundTag> nbt, final List<SupportingEntity> support, final Optional<SupportingEntity> mount, final List<Exclusion> exclusions, final boolean finalize) {
        this.weight = weight;
        this.quality = quality;
        this.chance = chance;
        this.name = name;
        this.entities = entities;
        this.stats = stats;
        this.stages = stages.orElse(null);
        this.dimensions = dimensions;
        this.affixed = affixed;
        this.gearSets = gearSets;
        this.nbt = nbt.orElse(null);
        this.support = support;
        this.mount = mount.orElse(null);
        this.exclusions = exclusions;
        this.finalize = finalize;
    }
    
    public int getWeight() {
        return this.weight;
    }
    
    public float getQuality() {
        return this.quality;
    }
    
    public float getChance() {
        return this.chance;
    }
    
    public Set<EntityType<?>> getEntities() {
        return this.entities;
    }
    
    public void transformMiniboss(final ServerLevelAccessor level, final Mob mob, final RandomSource random, final float luck) {
        final Vec3 pos = mob.m_20318_(0.0f);
        if (this.nbt != null && this.nbt.m_128441_("Passengers")) {
            final ListTag passengers = this.nbt.m_128437_("Passengers", 10);
            for (int i = 0; i < passengers.size(); ++i) {
                final Entity entity = EntityType.m_20645_(passengers.m_128728_(i), (Level)level.m_6018_(), (Function)Function.identity());
                if (entity != null) {
                    entity.m_7998_((Entity)mob, true);
                }
            }
        }
        mob.m_146884_(pos);
        this.initBoss(random, mob, luck);
        if (this.nbt != null) {
            mob.m_7378_(this.nbt);
        }
        if (this.mount != null) {
            final Mob mountedEntity = this.mount.create(mob.m_9236_(), mob.m_20185_() + 0.5, mob.m_20186_(), mob.m_20189_() + 0.5);
            mob.m_7998_((Entity)mountedEntity, true);
            level.m_7967_((Entity)mountedEntity);
        }
        if (this.support != null) {
            for (final SupportingEntity support : this.support) {
                final Mob supportingMob = support.create(mob.m_9236_(), mob.m_20185_() + 0.5, mob.m_20186_(), mob.m_20189_() + 0.5);
                level.m_7967_((Entity)supportingMob);
            }
        }
    }
    
    public void initBoss(final RandomSource rand, final Mob mob, final float luck) {
        mob.getPersistentData().m_128379_("apoth.miniboss", true);
        final int duration = (mob instanceof Creeper) ? 6000 : Integer.MAX_VALUE;
        for (final ChancedEffectInstance inst : this.stats.effects()) {
            if (rand.m_188501_() <= inst.chance()) {
                mob.m_7292_(inst.create(rand, duration));
            }
        }
        for (final RandomAttributeModifier modif : this.stats.modifiers()) {
            modif.apply(rand, (LivingEntity)mob);
        }
        if ("use_name_generation".equals(this.name)) {
            NameHelper.setEntityName(rand, mob);
        }
        else if (!Strings.isNullOrEmpty(this.name)) {
            mob.m_6593_((Component)Component.m_237115_(this.name));
        }
        if (mob.m_8077_()) {
            mob.m_20340_(true);
        }
        if (!this.gearSets.isEmpty()) {
            final GearSet set = GearSetRegistry.INSTANCE.getRandomSet(rand, luck, (List)this.gearSets);
            Preconditions.checkNotNull(set, (Object)String.format("Failed to find a valid gear set for the miniboss %s.", MinibossRegistry.INSTANCE.getKey((CodecProvider)this)));
            set.apply((LivingEntity)mob);
        }
        int guaranteed = -1;
        if (this.affixed) {
            boolean anyValid = false;
            for (final EquipmentSlot t : EquipmentSlot.values()) {
                final ItemStack s = mob.m_6844_(t);
                if (!s.m_41619_() && !LootCategory.forItem(s).isNone()) {
                    anyValid = true;
                    break;
                }
            }
            if (!anyValid) {
                AdventureModule.LOGGER.error("Attempted to affix a miniboss with ID " + MinibossRegistry.INSTANCE.getKey((CodecProvider)this) + " but it is not wearing any affixable items!");
                return;
            }
            ItemStack temp;
            for (guaranteed = rand.m_188503_(6), temp = mob.m_6844_(EquipmentSlot.values()[guaranteed]); temp.m_41619_() || LootCategory.forItem(temp) == LootCategory.NONE; temp = mob.m_6844_(EquipmentSlot.values()[guaranteed])) {
                guaranteed = rand.m_188503_(6);
            }
            final LootRarity rarity = LootRarity.random(rand, luck, (RarityClamp)AdventureConfig.AFFIX_CONVERT_RARITIES.get(mob.m_9236_().m_46472_().m_135782_()));
            ApothBoss.modifyBossItem(temp, rand, mob.m_8077_() ? mob.m_7770_().getString() : "", luck, rarity, this.stats);
            mob.m_6593_((Component)((MutableComponent)mob.m_7770_()).m_130948_(Style.f_131099_.m_131148_(rarity.getColor())));
            mob.m_21409_(EquipmentSlot.values()[guaranteed], 2.0f);
        }
        for (final EquipmentSlot s2 : EquipmentSlot.values()) {
            final ItemStack stack = mob.m_6844_(s2);
            if (!stack.m_41619_() && s2.ordinal() != guaranteed && rand.m_188501_() < this.stats.enchantChance()) {
                ApothBoss.enchantBossItem(rand, stack, Apotheosis.enableEnch ? this.stats.enchLevels()[0] : this.stats.enchLevels()[1], true);
                mob.m_8061_(s2, stack);
            }
        }
        mob.m_21153_(mob.m_21233_());
    }
    
    public ApothMiniboss validate(final ResourceLocation key) {
        Preconditions.checkArgument(this.weight >= 0, "Miniboss Item " + key + " has a negative weight!");
        Preconditions.checkArgument(this.quality >= 0.0f, "Miniboss Item " + key + " has a negative quality!");
        Preconditions.checkNotNull(this.entities, "Miniboss Item " + key + " has null entity match list!");
        Preconditions.checkNotNull(this.stats, "Miniboss Item " + key + " has no stats!");
        return this;
    }
    
    public Set<ResourceLocation> getDimensions() {
        return this.dimensions;
    }
    
    public Set<String> getStages() {
        return this.stages;
    }
    
    public Codec<? extends ApothMiniboss> getCodec() {
        return ApothMiniboss.CODEC;
    }
    
    public boolean requiresNbtAccess() {
        return this.exclusions.stream().anyMatch(Exclusion::requiresNbtAccess);
    }
    
    public boolean isExcluded(final Mob mob, final ServerLevelAccessor level, final MobSpawnType type) {
        final CompoundTag tag = this.requiresNbtAccess() ? mob.m_20240_(new CompoundTag()) : null;
        return this.exclusions.stream().anyMatch(ex -> ex.isExcluded(mob, level, type, tag));
    }
    
    public boolean shouldFinalize() {
        return this.finalize;
    }
    
    static {
        CODEC = RecordCodecBuilder.create(inst -> inst.group((App)Codec.intRange(0, Integer.MAX_VALUE).fieldOf("weight").forGetter((Function)WeightedDynamicRegistry.ILuckyWeighted::getWeight), (App)PlaceboCodecs.nullableField(Codec.floatRange(0.0f, Float.MAX_VALUE), "quality", (Object)0.0f).forGetter((Function)WeightedDynamicRegistry.ILuckyWeighted::getQuality), (App)ExtraCodecs.f_184349_.fieldOf("chance").forGetter(a -> a.chance), (App)PlaceboCodecs.nullableField((Codec)Codec.STRING, "name", (Object)"").forGetter(a -> a.name), (App)PlaceboCodecs.setOf(ForgeRegistries.ENTITY_TYPES.getCodec()).fieldOf("entities").forGetter(a -> a.entities), (App)BossStats.CODEC.fieldOf("stats").forGetter(a -> a.stats), (App)PlaceboCodecs.nullableField(PlaceboCodecs.setOf((Codec)Codec.STRING), "stages").forGetter(a -> Optional.ofNullable(a.stages)), (App)PlaceboCodecs.setOf(ResourceLocation.f_135803_).fieldOf("dimensions").forGetter(a -> a.dimensions), (App)PlaceboCodecs.nullableField((Codec)Codec.BOOL, "affixed", (Object)false).forGetter(a -> a.affixed), (App)PlaceboCodecs.nullableField(GearSet.SetPredicate.CODEC.listOf(), "valid_gear_sets", (Object)Collections.emptyList()).forGetter(a -> a.gearSets), (App)PlaceboCodecs.nullableField(NBTAdapter.EITHER_CODEC, "nbt").forGetter(a -> Optional.ofNullable(a.nbt)), (App)PlaceboCodecs.nullableField(SupportingEntity.CODEC.listOf(), "supporting_entities", (Object)Collections.emptyList()).forGetter(a -> a.support), (App)PlaceboCodecs.nullableField(SupportingEntity.CODEC, "mount").forGetter(a -> Optional.ofNullable(a.mount)), (App)PlaceboCodecs.nullableField(Exclusion.CODEC.listOf(), "exclusions", (Object)Collections.emptyList()).forGetter(a -> a.exclusions), (App)PlaceboCodecs.nullableField((Codec)Codec.BOOL, "finalize", (Object)false).forGetter(a -> a.finalize)).apply((Applicative)inst, ApothMiniboss::new));
    }
}
